/*
 *  Licensed to the Apache Software Foundation (ASF) under one or more
 *  contributor license agreements.  See the NOTICE file distributed with
 *  this work for additional information regarding copyright ownership.
 *  The ASF licenses this file to You under the Apache License, Version 2.0
 *  (the "License"); you may not use this file except in compliance with
 *  the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.coyote.http11.upgrade;

import java.io.IOException;
import java.io.OutputStream;
import java.net.Socket;

import org.apache.tomcat.util.net.SocketWrapper;

public class BioServletOutputStream extends AbstractServletOutputStream {

    private final OutputStream os;

    public BioServletOutputStream(SocketWrapper<Socket> socketWrapper,
            int asyncWriteBufferSize) throws IOException {
        super(asyncWriteBufferSize);
        os = socketWrapper.getSocket().getOutputStream();
    }

    @Override
    protected int doWrite(boolean block, byte[] b, int off, int len)
            throws IOException {
        os.write(b, off, len);
        return len;
    }

    @Override
    protected void doFlush() throws IOException {
        os.flush();
    }

    @Override
    protected void doClose() throws IOException {
        os.close();
    }
}
