/* Generated By:JavaCC: Do not edit this line. ELParserTreeConstants.java Version 5.0 */
package org.apache.el.parser;

public interface ELParserTreeConstants
{
  public int JJTCOMPOSITEEXPRESSION = 0;
  public int JJTLITERALEXPRESSION = 1;
  public int JJTDEFERREDEXPRESSION = 2;
  public int JJTDYNAMICEXPRESSION = 3;
  public int JJTVOID = 4;
  public int JJTCHOICE = 5;
  public int JJTOR = 6;
  public int JJTAND = 7;
  public int JJTEQUAL = 8;
  public int JJTNOTEQUAL = 9;
  public int JJTLESSTHAN = 10;
  public int JJTGREATERTHAN = 11;
  public int JJTLESSTHANEQUAL = 12;
  public int JJTGREATERTHANEQUAL = 13;
  public int JJTPLUS = 14;
  public int JJTMINUS = 15;
  public int JJTMULT = 16;
  public int JJTDIV = 17;
  public int JJTMOD = 18;
  public int JJTNEGATIVE = 19;
  public int JJTNOT = 20;
  public int JJTEMPTY = 21;
  public int JJTVALUE = 22;
  public int JJTDOTSUFFIX = 23;
  public int JJTBRACKETSUFFIX = 24;
  public int JJTMETHODPARAMETERS = 25;
  public int JJTIDENTIFIER = 26;
  public int JJTFUNCTION = 27;
  public int JJTTRUE = 28;
  public int JJTFALSE = 29;
  public int JJTFLOATINGPOINT = 30;
  public int JJTINTEGER = 31;
  public int JJTSTRING = 32;
  public int JJTNULL = 33;


  public String[] jjtNodeName = {
    "CompositeExpression",
    "LiteralExpression",
    "DeferredExpression",
    "DynamicExpression",
    "void",
    "Choice",
    "Or",
    "And",
    "Equal",
    "NotEqual",
    "LessThan",
    "GreaterThan",
    "LessThanEqual",
    "GreaterThanEqual",
    "Plus",
    "Minus",
    "Mult",
    "Div",
    "Mod",
    "Negative",
    "Not",
    "Empty",
    "Value",
    "DotSuffix",
    "BracketSuffix",
    "MethodParameters",
    "Identifier",
    "Function",
    "True",
    "False",
    "FloatingPoint",
    "Integer",
    "String",
    "Null",
  };
}
/* JavaCC - OriginalChecksum=437008e736f149e8fa6712fb36d831a1 (do not edit this line) */
